# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import textwrap

import pytest
from pyfakefs.fake_filesystem_unittest import Patcher

from ..state import DesignFormat
from ..config import Config, Macro


@pytest.fixture()
def example_config() -> Config:
    from ..config import Instance

    return Config(
        {
            "DEFAULT_CORNER": "nom_tt_025C_1v80",
            "LIB": {
                "*": "/pdk/my.lib",
            },
            "MACROS": {
                "a": Macro(
                    gds=[""],
                    lef=[""],
                    instances={"instance_a": Instance((0, 10), "N")},
                    nl=[],
                    spef={},
                    lib={
                        "nom_tt_025C_1v80": ["/cwd/a/lib/tt.lib"],
                        "nom_ss_n40C_1v80": ["/cwd/a/lib/ss.lib"],
                        "min_ff_025C_5v00": ["/cwd/a/lib/ff.lib"],
                    },
                    spice=[],
                    sdf={},
                    json_h=None,
                ),
                "b": Macro(
                    gds=[""],
                    lef=[""],
                    instances={"instance_b": Instance((0, 10), "FS")},
                    nl=[
                        "/cwd/b/spef/nl1.v",
                        "/cwd/b/spef/nl2.v",
                    ],
                    spef={
                        "min_*": ["/cwd/b/spef/min.spef"],
                        "nom_*": ["/cwd/b/spef/nom.spef"],
                        "max_*": ["/cwd/b/spef/max.spef"],
                    },
                    lib={
                        "nom_tt_025C_1v80": ["/cwd/b/lib/tt.lib"],
                        "nom_ss_n40C_1v80": ["/cwd/b/lib/ss.lib"],
                        "min_ff_025C_5v00": ["/cwd/b/lib/ff.lib"],
                    },
                    spice=[],
                    sdf={},
                    json_h=None,
                ),
            },
        }
    )


@pytest.mark.parametrize(
    ("corner", "expected"),
    [
        (None, ["file1.spef", "typical.whatever"]),
        ("nom_ss_n40C_1v80", ["file1.spef", "ss.a", "🥶"]),
        ("min_ff_025C_5v00", []),
    ],
)
def test_filter_views(corner, expected, example_config: Config):
    from . import Toolbox
    from ..common import Path

    toolbox = Toolbox(".")
    views_by_corner = {
        k: Path(v)
        for k, v in {
            "nom_*": "file1.spef",
            "*_tt_*": "typical.whatever",
            "*_ss_*": "ss.a",
            "nom_": "no globs?",
            "*_100C_*": "🔥",
            "*_n40C_*": "🥶",
        }.items()
    }

    assert toolbox.filter_views(example_config, views_by_corner, corner) == expected, []


def gmv_parameters(f):
    f = pytest.mark.parametrize(
        ("view", "corner", "unless_exist", "expected"),
        [
            (DesignFormat.LIB, None, None, ["/cwd/a/lib/tt.lib", "/cwd/b/lib/tt.lib"]),
            (DesignFormat.LIB, None, DesignFormat.SPEF, ["/cwd/a/lib/tt.lib"]),
            (DesignFormat.SPEF, "min_ff_025C_5V0", None, ["/cwd/b/spef/min.spef"]),
            (DesignFormat.SPEF, "not_a_real_corner", None, []),
            (
                DesignFormat.NETLIST,
                "nom_ss_n40C_1v80",
                None,
                [
                    "/cwd/b/spef/nl1.v",
                    "/cwd/b/spef/nl2.v",
                ],
            ),
            (
                DesignFormat.LIB,
                "nom_ss_n40C_1v80",
                DesignFormat.NETLIST,
                ["/cwd/a/lib/ss.lib"],
            ),
            (
                DesignFormat.SDC,
                "nom_ss_n40C_1v80",
                None,
                [],
            ),
        ],
    )(f)
    return f


@gmv_parameters
def test_get_macro_views_without_macros(
    view,
    corner,
    unless_exist,
    expected,
):
    from . import Toolbox

    toolbox = Toolbox(".")

    assert (
        toolbox.get_macro_views(
            Config({"DEFAULT_CORNER": "nom_ss_n40C_1v80", "MACROS": None}),
            view,
            corner,
            unless_exist,
        )
        == []
    ), "get_macro_views with empty macro object returned a non-empty list"


@gmv_parameters
def test_get_macro_views_with_macros(
    view,
    corner,
    unless_exist,
    expected,
    example_config,
):
    from . import Toolbox

    toolbox = Toolbox(".")

    assert (
        toolbox.get_macro_views(example_config, view, corner, unless_exist) == expected
    ), "get_macro_views returned unexpected result"


@pytest.mark.parametrize(
    (
        "timing_corner",
        "prioritize_nl",
        "expected",
    ),
    [
        (
            None,
            False,
            (
                "nom_tt_025C_1v80",
                ["/pdk/my.lib", "/cwd/a/lib/tt.lib", "/cwd/b/lib/tt.lib"],
            ),
        ),
        (
            None,
            True,
            (
                "nom_tt_025C_1v80",
                [
                    "/pdk/my.lib",
                    "/cwd/a/lib/tt.lib",
                    "/cwd/b/spef/nl1.v",
                    "/cwd/b/spef/nl2.v",
                    "instance_b@/cwd/b/spef/nom.spef",
                ],
            ),
        ),
        (
            "min_ff_025C_5v00",
            True,
            (
                "min_ff_025C_5v00",
                [
                    "/pdk/my.lib",
                    "/cwd/a/lib/ff.lib",
                    "/cwd/b/spef/nl1.v",
                    "/cwd/b/spef/nl2.v",
                    "instance_b@/cwd/b/spef/min.spef",
                ],
            ),
        ),
    ],
)
def test_get_timing_files(timing_corner, prioritize_nl, expected, example_config):
    from . import Toolbox

    toolbox = Toolbox(".")

    assert (
        toolbox.get_timing_files(
            example_config,
            timing_corner,
            prioritize_nl,
        )
        == expected
    ), "get_timing_files returned unexpected result"


def test_get_timing_files_warnings(
    caplog: pytest.LogCaptureFixture, example_config: Config
):
    from . import Toolbox

    toolbox = Toolbox(".")

    cfg = example_config

    # 0. Missing netlists
    netlist_bk = cfg["MACROS"]["b"].nl
    cfg["MACROS"]["b"].nl = []

    assert toolbox.get_timing_files(cfg, None, True,) == (
        "nom_tt_025C_1v80",
        ["/pdk/my.lib", "/cwd/a/lib/tt.lib", "/cwd/b/lib/tt.lib"],
    ), "get_timing_files returned unexpected result"

    assert (
        "but no netlist found" in caplog.text
    ), "get_timing_files did not warn about missing netlists"

    cfg["MACROS"]["b"].nl = netlist_bk

    # 1. Missing spefs
    spefs_bk = cfg["MACROS"]["b"].spef
    cfg["MACROS"]["b"].spef = {}

    assert toolbox.get_timing_files(cfg, None, True,) == (
        "nom_tt_025C_1v80",
        ["/pdk/my.lib", "/cwd/a/lib/tt.lib", "/cwd/b/lib/tt.lib"],
    ), "get_timing_files returned unexpected result"

    assert (
        "but no parasitics extraction found" in caplog.text
    ), "get_timing_files did not warn about missing spefs"

    cfg["MACROS"]["b"].spef = spefs_bk

    # 2. No SCLs
    cfg = cfg.copy(LIB={})

    assert toolbox.get_timing_files(cfg, None, True,) == (
        "nom_tt_025C_1v80",
        [
            "/cwd/a/lib/tt.lib",
            "/cwd/b/spef/nl1.v",
            "/cwd/b/spef/nl2.v",
            "instance_b@/cwd/b/spef/nom.spef",
        ],
    ), "get_timing_files returned unexpected result"

    assert (
        "No SCL lib files" in caplog.text
    ), "get_timing_files did not warn about missing SCL liberty"


@pytest.fixture()
def _mock_fs():
    with Patcher() as patcher:
        patcher.fs.create_dir("/cwd")
        os.chdir("/cwd")
        patcher.fs.create_file(
            "/cwd/example_lib.lib",
            contents=textwrap.dedent(
                """
                library ("example_lib") {
                    cell ("example_lib__cell0") {
                    }
                    cell ("example_lib__cell1") {
                        property (x) {
                            
                        }
                    }
                    cell ("example_lib__cell2") {
                    }
                }
                """
            ),
        )
        patcher.fs.create_file(
            "/cwd/example_lib2.lib",
            contents=textwrap.dedent(
                """
                library ("example_lib2") {
                    cell ("example_lib2__cell0") {
                    }
                    cell ("example_lib2__cell1") {
                    }
                    cell ("example_lib2__cell2") {
                    }
                }
                """
            ),
        )
        patcher.fs.create_file(
            "/cwd/bad_cell_list.txt",
            contents=textwrap.dedent(
                """
                example_lib__cell0
                example_lib2__cell1
                example_lib2__cell2
                """
            ),
        )
        yield


@pytest.fixture()
def lib_trim_result():
    return [
        textwrap.dedent(
            """
                library ("example_lib") {
                    /* removed example_lib__cell0 */
                    cell ("example_lib__cell1") {
                        property (x) {

                        }
                    }
                    cell ("example_lib__cell2") {
                    }
                }
                """
        ).strip(),
        textwrap.dedent(
            """
                library ("example_lib2") {
                    cell ("example_lib2__cell0") {
                    }
                    /* removed example_lib2__cell1 */
                    /* removed example_lib2__cell2 */
                }
                """
        ).strip(),
    ]


@pytest.mark.usefixtures("_mock_fs")
def test_remove_cell_list_from_lib(lib_trim_result):
    from . import Toolbox

    toolbox = Toolbox(".")

    result = toolbox.remove_cells_from_lib(
        frozenset(["/cwd/example_lib.lib", "/cwd/example_lib2.lib"]),
        excluded_cells=frozenset(["/cwd/bad_cell_list.txt"]),
        as_cell_lists=True,
    )
    for file in result:
        contents = open(file, encoding="utf8").read()
        assert (
            contents.strip() in lib_trim_result
        ), "remove_cells_from_lib produced unexpected result"


@pytest.mark.usefixtures("_mock_fs")
def test_remove_cells_from_lib(lib_trim_result):
    from . import Toolbox

    toolbox = Toolbox(".")

    excluded_cells = (
        open("/cwd/bad_cell_list.txt", encoding="utf8").read().strip().splitlines()
    )

    result = toolbox.remove_cells_from_lib(
        frozenset(["/cwd/example_lib.lib", "/cwd/example_lib2.lib"]),
        excluded_cells=frozenset(excluded_cells),
    )
    for file in result:
        contents = open(file, encoding="utf8").read()
        assert (
            contents.strip() in lib_trim_result
        ), "remove_cells_from_lib produced unexpected result"



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import textwrap
from typing import Tuple

import pytest

from .step import Step, ViewsUpdate, MetricsUpdate
from ..state import State
from ..config import Config
from ..config.test_config import _mock_fs, mock_variables  # noqa: F401


class chdir(object):
    def __init__(self, path):
        self.path = path
        self.previous = None

    def __enter__(self):
        self.previous = os.getcwd()
        os.chdir(self.path)

    def __exit__(self, exc_type, exc_value, traceback):
        os.chdir(self.previous)
        if exc_type is not None:
            raise exc_value


@pytest.fixture()
def mock_run():
    def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
        views_update: ViewsUpdate = {}
        metrics: MetricsUpdate = {}
        return views_update, metrics

    return run


def test_step_init_empty():
    with pytest.raises(TypeError, match="Can't instantiate abstract class Step"):
        Step()


def test_step_init_missing_id(mock_run):
    class TestStep(Step):
        inputs = []
        outputs = []
        run = mock_run

    with pytest.raises(
        NotImplementedError,
        match="All subclasses of Step must override the value of id",
    ):
        TestStep()


def test_step_inputs_not_implemented(mock_run):
    with pytest.raises(
        NotImplementedError, match="must implement class attribute 'inputs'"
    ):

        class TestStep(Step):
            run = mock_run


def test_step_outputs_not_implemented(mock_run):
    with pytest.raises(
        NotImplementedError, match="must implement class attribute 'outputs'"
    ):

        class TestStep(Step):
            inputs = []
            run = mock_run


def test_step_missing_config(mock_run):
    class TestStep(Step):
        inputs = []
        outputs = []
        run = mock_run
        id = "TestStep"

    with pytest.raises(TypeError, match="Missing required argument 'config'"):
        TestStep()


def test_step_missing_state_in(mock_run):
    class TestStep(Step):
        inputs = []
        outputs = []
        run = mock_run
        id = "TestStep"

    with pytest.raises(TypeError, match="Missing required argument 'state_in'"):

        TestStep(config=Config({"abc": "abc"}))


@pytest.fixture()
@mock_variables()
def mock_config():
    from ..config import config

    mock_config, _ = Config.load(
        {
            "DESIGN_NAME": "whatever",
            "VERILOG_FILES": "dir::src/*.v",
        },
        config.flow_common_variables,
        design_dir="/cwd",
        pdk="dummy",
        scl="dummy_scl",
        pdk_root="/pdk",
    )
    return mock_config


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_step_create(mock_run, mock_config):
    from ..state.design_format import DesignFormat

    class TestStep(Step):
        inputs = []
        outputs = []
        run = mock_run
        id = "TestStep"

    step = TestStep(
        id="TestStep",
        long_name="longname",
        config=mock_config,
        state_in=State({DesignFormat.NETLIST: "abc"}),
    )
    assert step.id == "TestStep", "Wrong step id"
    assert step.long_name == "longname", "Wrong step longname"
    assert step.config == mock_config, "Wrong step config"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_mock_run(mock_run, mock_config):
    from ..state.design_format import DesignFormat

    class TestStep(Step):
        inputs = []
        outputs = []
        run = mock_run
        id = "TestStep"

    state_in = State({DesignFormat.NETLIST: "abc"})
    step = TestStep(
        id="TestStep",
        long_name="longname",
        config=mock_config,
        state_in=state_in,
    )
    views_update, metrics_update = step.run(state_in)
    assert step.id == "TestStep", "Wrong step id"
    assert (
        step.long_name == "longname"
    ), "Wrong step long_name, declared via constructor"
    assert step.config == mock_config, "Wrong step config"
    assert views_update == {}, "Wrong step run -- tainted views_update"
    assert metrics_update == {}, "Wrong step run -- tainted metrics_update"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_step_start_missing_toolbox(mock_run, mock_config):
    from ..state.design_format import DesignFormat

    class TestStep(Step):
        inputs = []
        outputs = []
        run = mock_run
        id = "TestStep"

    state_in = State({DesignFormat.NETLIST: "abc"})
    step = TestStep(
        id="TestStep",
        long_name="longname",
        config=mock_config,
        state_in=state_in,
    )
    with pytest.raises(TypeError, match="Missing argument 'toolbox' required"):
        step.start()


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_step_start_missing_step_dir(mock_run, mock_config):
    from ..state.design_format import DesignFormat
    from ..utils import Toolbox

    class TestStep(Step):
        inputs = []
        outputs = []
        run = mock_run
        id = "TestStep"

    state_in = State({DesignFormat.NETLIST: "abc"})
    step = TestStep(
        id="TestStep",
        long_name="longname",
        config=mock_config,
        state_in=state_in,
    )
    with pytest.raises(TypeError, match="Missing required argument 'step_dir'"):
        step.start(toolbox=Toolbox(tmp_dir="/cwd"))


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_step_start_invalid_state(mock_run, mock_config):
    from ..state.design_format import DesignFormat
    from ..utils import Toolbox
    from .step import StepException

    class TestStep(Step):
        inputs = []
        outputs = []
        run = mock_run
        id = "TestStep"

    state_in = State({DesignFormat.NETLIST: "abc"})
    step = TestStep(
        id="TestStep",
        long_name="longname",
        config=mock_config,
        state_in=state_in,
    )
    with pytest.raises(StepException, match="generated invalid state"):
        step.start(toolbox=Toolbox(tmp_dir="/cwd"), step_dir="/cwd")


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_step_start(mock_config):
    from ..common import Path
    from ..state.design_format import DesignFormat
    from ..utils import Toolbox

    test_file = "test.nl.v"
    with open(test_file, "w") as f:
        f.write("\n")

    test_file_out = "test2.nl.v"
    with open(test_file_out, "w") as f:
        f.write("\n")

    new_metric: MetricsUpdate = {"new_metric": "abc"}
    new_view: ViewsUpdate = {DesignFormat.NETLIST: Path(test_file_out)}

    class TestStep(Step):
        inputs = []
        outputs = []
        id = "TestStep"

        def run(self, state_in: State, **kwargs) -> Tuple[ViewsUpdate, MetricsUpdate]:
            return new_view, new_metric

    metrics_in = {"metric": "123"}
    state_in = State({DesignFormat.NETLIST: Path(test_file)}, metrics=metrics_in)
    step = TestStep(
        id="TestStep",
        long_name="longname",
        config=mock_config,
        state_in=state_in,
    )
    state_out = step.start(toolbox=Toolbox(tmp_dir="/cwd"), step_dir="/cwd")
    assert state_out[DesignFormat.NETLIST] == test_file_out, "Wrong step state_out"
    assert state_out.metrics == {
        **new_metric,
        **metrics_in,
    }, "Wrong step state_out metrics"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_step_longname(mock_run, mock_config):
    from ..state.design_format import DesignFormat

    class TestStep(Step):
        inputs = []
        outputs = []
        run = mock_run
        id = "TestStep"
        long_name = "longname2"

    step = TestStep(
        id="TestStep",
        config=mock_config,
        state_in=State({DesignFormat.NETLIST: "abc"}),
    )
    assert step.long_name == "longname2", "Wrong long_name declared via class"


@pytest.mark.usefixtures("_mock_fs")
@mock_variables()
def test_step_factory(mock_run):
    @Step.factory.register()
    class TestStep(Step):
        inputs = []
        outputs = []
        run = mock_run
        id = "TestStep"
        long_name = "longname2"

    assert (
        "TestStep" in Step.factory.list()
    ), "Step factory did not register step used for testing: TestStep"
    assert (
        Step.factory.get("TestStep") == TestStep
    ), "Wrong type registered by StepFactor"


@mock_variables()
def test_run_subprocess(mock_run):
    import tempfile
    import subprocess

    from ..config.config import Config
    from ..config.config import Meta
    from ..state.design_format import DesignFormat

    with tempfile.TemporaryDirectory() as dir, chdir(dir):

        state_in = State({DesignFormat.NETLIST: "abc"})

        class StepTest(Step):
            inputs = [DesignFormat.NETLIST]
            outputs = [DesignFormat.NETLIST]
            id = "TclStepTest"
            step_dir = dir
            run = mock_run

        config_dict = {
            "DESIGN_NAME": "whatever",
            "DESIGN_DIR": dir,
            "EXAMPLE_PDK_VAR": "bla",
            "PDK": "dummy",
            "STD_CELL_LIBRARY": "dummy_scl",
            "VERILOG_FILES": ["/cwd/src/a.v", "/cwd/src/b.v"],
            "GRT_REPAIR_ANTENNAS": True,
            "RUN_HEURISTIC_DIODE_INSERTION": False,
            "MACROS": None,
            "DIODE_ON_PORTS": None,
            "TECH_LEFS": {
                "nom_*": "/pdk/dummy/libs.ref/techlef/dummy_scl/dummy_tech_lef.tlef"
            },
            "DEFAULT_CORNER": "nom_tt_025C_1v80",
        }
        step = StepTest(
            config=Config(config_dict, meta=Meta(version=2, flow="n/a")),
            state_in=state_in,
        )
        out_file = "out.txt"
        report_file = "test.rpt"
        report_data = "Hello World"
        extra_data = "Bye World"
        new_metric = {"new_metric": 1, "new_float_metric": 2.0}
        out_data = textwrap.dedent(
            f"""
            %OL_CREATE_REPORT {report_file}
            {report_data}
            %OL_END_REPORT
            {extra_data}
            %OL_METRIC_I new_metric {new_metric['new_metric']}
            %OL_METRIC_F new_float_metric {new_metric['new_float_metric']}
            """
        ).strip()

        with open(out_file, "w") as f:
            f.write(out_data)

        subprocess_log_file = "test.log"
        out_metrics = step.run_subprocess(
            ["cat", out_file], silent=True, log_to=subprocess_log_file
        )
        actual_out_data = ""
        with open(subprocess_log_file) as f:
            actual_out_data = f.read()
        actual_report_data = ""
        with open(report_file) as f:
            actual_report_data = f.read()

        assert out_metrics == new_metric, ".run_subprocess() generated invalid metrics"
        assert (
            actual_report_data.strip() == report_data
        ), ".run_subprocess() generated invalid report"
        assert (
            actual_out_data == out_data
        ), ".run_subprocess() generated mis-matched log file"

    with pytest.raises(subprocess.CalledProcessError):
        step.run_subprocess(["false"])
